# -*- coding: iso-8859-1 -*-
"""
    MoinMoin - MoinForms library

    @copyright: 2012 by Paul Boddie <paul@boddie.org.uk>
    @license: GNU GPL (v2 or later), see COPYING.txt for details.
"""

from MoinMoin import wikiutil
from MoinSupport import *
import re

__version__ = "0.1"

form_field_regexp_str = r"<<FormField\((.*?)\)>>"
form_field_regexp = re.compile(form_field_regexp_str, re.DOTALL)

# Common formatting functions.

def formatForm(text, request, fmt, attrs=None, write=None):

    """
    Format the given 'text' using the specified 'request' and formatter 'fmt'.
    The optional 'attrs' can be used to control the presentation of the form.

    If the 'write' parameter is specified, use it to write output; otherwise,
    write output using the request.
    """

    write = write or request.write
    page = request.page

    fields = getFields(get_form(request))

    queryparams = []

    for argname in ["fragment", "action"]:
        if attrs and attrs.has_key(argname):
            queryparams.append("%s=%s" % (argname, attrs[argname]))

    querystr = "&".join(queryparams)

    write(fmt.rawHTML('<form method="post" action="%s">' %
        escattr(page.url(request, querystr))
        ))

    output = getFormOutput(text, fields)
    write(formatText(output, request, fmt, inhibit_p=False))

    write(fmt.rawHTML('</form>'))

def getFormOutput(text, fields):

    """
    Combine regions found in the given 'text' and then return them as a single
    block. The reason for doing this, as opposed to just passing each region to
    a suitable parser for formatting, is that form sections may break up
    regions, and such sections may not define separate subregions but instead
    act as a means of conditional inclusion of text into an outer region.

    The given 'fields' are used to populate fields provided in forms and to
    control whether sections are populated or not.
    """

    output = []
    section = fields

    for region in getRegions(text, True):
        format, attributes, body, header, close = getFragmentFromRegion(region)

        # Include bare regions as they are.

        if format is None:
            output.append(region)

        # Include form sections only if fields exist for those sections.

        elif format == "form":
            section_name = attributes.get("section")
            if section_name and section.has_key(section_name):

                # Iterate over the section contents ignoring the given indexes.

                for index, element in enumerate(getSectionElements(section[section_name])):

                    # Adjust FormField macros to use hierarchical names.

                    adjusted_body = adjustFormFields(body, section_name, index)
                    output.append(getFormOutput(adjusted_body, element))

        # Inspect and include other regions.

        else:
            output.append(header)
            output.append(getFormOutput(body, section))
            output.append(close)

    return "".join(output)

def adjustFormFields(body, section_name, index):

    """
    Return a version of the 'body' with the names in FormField macros updated to
    incorporate the given 'section_name' and 'index'.
    """

    result = []
    in_macro = False

    for match in form_field_regexp.split(body):
        if not in_macro:
            result.append(match)
        else:
            result.append("<<FormField(%s)>>" % ",".join(
                adjustMacroArguments(parseMacroArguments(match), section_name, index)
                ))

        in_macro = not in_macro

    return "".join(result)

def adjustMacroArguments(args, section_name, index):

    """
    Adjust the given 'args' so that the path incorporates the given
    'section_name' and 'index', returning a new list containing the revised
    path and remaining arguments.
    """

    result = []
    path = None

    for arg in args:
        if arg.startswith("path="):
            path = arg[5:]
        else:
            result.append(arg)

    qualified = "%s%s$%s" % (path and ("%s/" % path) or "", section_name, index)
    result.append("path=%s" % qualified)

    return result

def parseMacroArguments(args):

    """
    Interpret the arguments.
    NOTE: The argument parsing should really be more powerful in order to
    NOTE: support labels.
    """

    try:
        parsed_args = args and wikiutil.parse_quoted_separated(args, name_value=False) or []
    except AttributeError:
        parsed_args = args.split(",")

    return [arg for arg in parsed_args if arg]

def getFields(d, remove=False):

    """
    Return the form fields hierarchy for the given dictionary 'd'. If the
    optional 'remove' parameter is set to a true value, remove the entries for
    the fields from 'd'.
    """

    fields = {}

    for key, value in d.items():

        # Detect modifying fields.

        if key.find("=") != -1:
            fields[key] = value
            if remove:
                del d[key]
            continue

        # Reproduce the original hierarchy of the fields.

        section = fields
        parts = getPathDetails(key)

        for name, index in parts[:-1]:

            # Add an entry for instances of the section.

            if not section.has_key(name):
                section[name] = {}

            # Add an entry for the specific instance of the section.

            if not section[name].has_key(index):
                section[name][index] = {}

            section = section[name][index]

        section[parts[-1][0]] = value

        if remove:
            del d[key]

    return fields

def getPathDetails(path):
    parts = []

    for part in path.split("/"):
        try:
            name, index = part.split("$", 1)
            index = int(index)
        except ValueError:
            name, index = part, None

        parts.append((name, index))

    return parts

def getSectionForPath(path, fields):

    """
    Obtain the section indicated by the given 'path' from the 'fields',
    returning a tuple of the form (parent section, (name, index)), where the
    parent section contains the referenced section, where name is the name of
    the referenced section, and where index, if not None, is the index of a
    specific section instance within the named section.
    """

    parts = getPathDetails(path)
    section = fields

    for name, index in parts[:-1]:
        section = fields[name][index]

    return section, parts[-1]

def getSectionElements(section_elements):

    "Return the given 'section_elements' as an ordered collection."

    keys = map(int, section_elements.keys())
    keys.sort()

    elements = []

    for key in keys:
        elements.append(section_elements[key])

    return elements

def formatFormForOutputType(text, request, mimetype, attrs=None, write=None):

    """
    Format the given 'text' using the specified 'request' for the given output
    'mimetype'.

    The optional 'attrs' can be used to control the presentation of the form.

    If the 'write' parameter is specified, use it to write output; otherwise,
    write output using the request.
    """

    write = write or request.write

    if mimetype == "text/html":
        write('<html>')
        write('<body>')
        fmt = request.html_formatter
        fmt.setPage(request.page)
        formatForm(text, request, fmt, attrs, write)
        write('</body>')
        write('</html>')

# vim: tabstop=4 expandtab shiftwidth=4
